/**
 * Copyright 2005 Jason LaDere
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Originally based on Project64 code.
 *
 */

/*
 * JglN64.java
 *
 * Created on November 26, 2005, 12:27 PM
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package gln64j;

import debug.Debug;
import gln64j.rsp.Gsp;
import gln64j.rdp.Gdp;
import java.nio.ByteBuffer;
import javax.swing.JFrame;
import plugin.GfxPlugin;

/**
 *
 * @author Jason
 */
public class GLN64jPlugin implements GfxPlugin {
    
    public static final boolean DEBUG = Debug.DEBUG_GLN64;
    
    private static final int NUM_FRAMES = 7;
    
//    public static GL gl;
    
    // used by OpenGl
    public static JFrame hWnd;
    
    // used by Gdp
    public static Runnable CheckInterrupts;
    
    public static Registers REG = new Registers();
    public static ByteBuffer DMEM;
    public static ByteBuffer IMEM;
    public static ByteBuffer TMEM = ByteBuffer.allocate(8 * 512);
    public static ByteBuffer RDRAM;
    public static int RDRAMSize;
    
    // used by OpenGl, Gdp
//    public static int screenWidth;
//    public static int screenHeight;
    
//    private long lastOrigin;
    private String name;
    
//    private long lastFrame;
//    private long[] frames = new long[NUM_FRAMES];
//    private int currentFrame;
//    private long frequency;
//    private int oldViVsyncReg = 0;
//    private int viIntrTime = 500000;
//    private int viFieldNumber;
    
    /** Creates a new instance of JglN64 */
    public GLN64jPlugin() {
//        frequency = 1000;
//        currentFrame = 0;
//        viFieldNumber = 0;
        
        name = this.getClass().getName();
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") created.");
        
        OpenGl.config();
    }
    
    // called by OpenGl
//    public static void VI_UpdateSize() {
//        float xScale = (REG.VI_Registers[REG.VI_X_SCALE]&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
//        float xOffset = ((REG.VI_Registers[REG.VI_X_SCALE]>>16)&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
//        
//        float yScale = (REG.VI_Registers[REG.VI_Y_SCALE]&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
//        float yOffset = ((REG.VI_Registers[REG.VI_X_SCALE]>>16)&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
//        
//        int hEnd = REG.VI_Registers[REG.VI_H_START]&Gbi.SR_MASK_10;
//        int hStart = (REG.VI_Registers[REG.VI_H_START]>>16)&Gbi.SR_MASK_10;
//        
//        // These are in half-lines, so shift an extra bit
//        int vEnd = (REG.VI_Registers[REG.VI_V_START]>>1)&Gbi.SR_MASK_9;
//        int vStart = (REG.VI_Registers[REG.VI_V_START]>>17)&Gbi.SR_MASK_9;
//        
//        OpenGlGdp.screenWidth = (int)((hEnd - hStart) * xScale);
//        OpenGlGdp.screenHeight = (int)((vEnd - vStart) * yScale * 1.0126582f);
//        
//        if (OpenGlGdp.screenWidth == 0)
//            OpenGlGdp.screenWidth = 320;
//        if (OpenGlGdp.screenHeight == 0)
//            OpenGlGdp.screenHeight = 240;
//    }
    
    public void captureScreen(String directory) {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") captureScreen.");
    }
    
    public void changeWindow() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") changeWindow.");
    }
    
    public void closePlugin() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") closePlugin.");
    }
    
    public void pluginAbout(JFrame hParent) {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") pluginAbout.");
    }
    
    public void pluginConfig(JFrame hParent) {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") pluginConfig.");
    }
    
    public void pluginTest(JFrame hParent) {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") pluginTest.");
    }
    
    public void drawScreen() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") drawScreen.");
    }
    
    public void getPluginInfo(PluginInfo pluginInfo) {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") getPluginInfo.");
    }
    
    public boolean initiateGfx(GfxInfo gfxInfo) {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") initiateGfx.");
        
        hWnd = gfxInfo.hWnd;

        DMEM = gfxInfo.dmem;
        IMEM = gfxInfo.imem;
        RDRAM = gfxInfo.rdram;
        
        REG.MI_INTR = gfxInfo.MI_INTR_REG;
        REG.MI_Registers = gfxInfo.miRegisters;
        
        REG.DPC_START = gfxInfo.DPC_START_REG;
        REG.DPC_END = gfxInfo.DPC_END_REG;
        REG.DPC_CURRENT = gfxInfo.DPC_CURRENT_REG;
        REG.DPC_STATUS = gfxInfo.DPC_STATUS_REG;
        REG.DPC_CLOCK = gfxInfo.DPC_CLOCK_REG;
        REG.DPC_BUFBUSY = gfxInfo.DPC_BUFBUSY_REG;
        REG.DPC_PIPEBUSY = gfxInfo.DPC_PIPEBUSY_REG;
        REG.DPC_TMEM = gfxInfo.DPC_TMEM_REG;
        REG.DPC_Registers = gfxInfo.dpcRegisters;
        
        REG.VI_STATUS = VI_STATUS_REG;
        REG.VI_ORIGIN = VI_ORIGIN_REG;
        REG.VI_WIDTH = VI_WIDTH_REG;
        REG.VI_INTR = VI_INTR_REG;
        REG.VI_V_CURRENT_LINE = VI_CURRENT_REG;
        REG.VI_TIMING = VI_BURST_REG;
        REG.VI_V_SYNC = VI_V_SYNC_REG;
        REG.VI_H_SYNC = VI_H_SYNC_REG;
        REG.VI_LEAP = VI_LEAP_REG;
        REG.VI_H_START = VI_H_START_REG;
        REG.VI_V_START = VI_V_START_REG;
        REG.VI_V_BURST = VI_V_BURST_REG;
        REG.VI_X_SCALE = VI_X_SCALE_REG;
        REG.VI_Y_SCALE = VI_Y_SCALE_REG;
        REG.VI_Registers = gfxInfo.viRegisters;
        
        CheckInterrupts = gfxInfo.checkInterrupts;
        
        return true;
    }
    
    public void moveScreen(int xpos, int ypos) {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") moveScreen.");
    }
    
    public void processDList() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") processDList.");
        OpenGlGdp.hDC.display();
    }
    
    public void processRDPList() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") processRDPList.");
    }
    
    public void romClosed() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") romClosed.");
        OpenGl.OGL_Stop();
        if (DEBUG) Debug.CloseDebugDlg();
    }
    
    public void romOpen() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") romOpen.");
        OpenGl.uc_start = OpenGl.uc_dstart = 0;
        RDRAMSize = RDRAM.capacity();
//        Rsp.RSP_Init();
        Rsp.gsp = new Gsp(RDRAM, DMEM);
        Rsp.gdp = new Gdp(CheckInterrupts, REG);
//        Gbi.gbi.init();
        OpenGl.OGL_Start();
        
        OpenGlGdp.OGL_ResizeWindow();
        if (DEBUG) Debug.OpenDebugDlg();
    }
    
    public void showCFB() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") showCFB.");
    }
    
    public void updateScreen() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") updateScreen.");
        //viUpdateScreen();
        
//        if (oldViVsyncReg != REG.VI_Registers[VI_V_SYNC_REG]) {
//            if (REG.VI_Registers[VI_V_SYNC_REG] == 0) {
//                viIntrTime = 500000;
//            } else {
//                viIntrTime = (REG.VI_Registers[VI_V_SYNC_REG] + 1) * 1500;
//                if ((REG.VI_Registers[VI_V_SYNC_REG] % 1) != 0) {
//                    viIntrTime -= 38;
//                }
//            }
//        }
//        if ((REG.VI_Registers[VI_STATUS_REG] & 0x10) != 0) {
//            if (viFieldNumber == 0) {
//                viFieldNumber = 1;
//            } else {
//                viFieldNumber = 0;
//            }
//        } else {
//            viFieldNumber = 0;
//        }
        
//        if ((currentFrame & 7) == 0) {
//            long time = System.currentTimeMillis();
//            frames[(currentFrame >> 3) % NUM_FRAMES] = time - lastFrame;
//            lastFrame = time;
//            if (currentFrame > (NUM_FRAMES << 3)) {
//                long total = 0;
//                for (int count=0; count<NUM_FRAMES; count++)
//                    total += frames[count];
//                hWnd.setTitle(String.format("FPS: %.2f", frequency / ((double)total / (NUM_FRAMES << 3))));
//            } else {
//                hWnd.setTitle("FPS: -.--");
//            }
//        }
//        currentFrame += 1;
        
        OpenGlGdp.viUpdateScreen();
        
//        return viIntrTime;
    }
    
    public void viStatusChanged() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") viStatusChanged.");
    }
    
    public void viWidthChanged() {
        if (DEBUG) System.out.println("GFX Plugin (" + name + ") viWidthChanged.");
    }
    
//    public int updateCurrentHalfLine(int timer) {
//        if (timer < 0)
//            return 0;
//        int halfLine = (timer / 1500);
//        halfLine &= ~1;
//        halfLine += viFieldNumber;
//        return halfLine;
//    }
    
    // Private Methods /////////////////////////////////////////////////////////
    
//    private void viUpdateScreen() {
//        OpenGl.gl.glFinish();
//        if (OpenGl.frameBufferTextures) {
//            FrameBufferStack.FrameBuffer buffer = OpenGl.frameBuffers.findBuffer(REG.VI_ORIGIN);
//            if ((REG.VI_ORIGIN != lastOrigin) || ((buffer != null) && buffer.changed)) {
//                if (Rsp.gdp.colorImage.changed) {
//                    OpenGl.saveBuffer(Rsp.gdp.colorImage.address, Rsp.gdp.colorImage.size, Rsp.gdp.colorImage.width, Rsp.gdp.colorImage.height);
//                    Rsp.gdp.changed |= Gdp.CHANGED_TEXTURE;
//                    Rsp.gdp.colorImage.changed = false;
//                }
//                if (OpenGl.renderBuffer(REG.VI_ORIGIN)) {
//                    Rsp.gsp.changed |= Gsp.CHANGED_VIEWPORT;
//                    Rsp.gdp.changed |= Gdp.CHANGED_TEXTURE;
//                    Rsp.gdp.changed |= Gdp.CHANGED_COMBINE;
//                }
//                Rsp.gdp.colorImage.changed = false;
//                lastOrigin = REG.VI_ORIGIN;
//            }
//        }
//        else {
//            if ((Rsp.gsp.changed & Gsp.CHANGED_COLORBUFFER) != 0) {
//                OpenGl.hDC.swapBuffers();
//                Rsp.gsp.changed &= ~Gsp.CHANGED_COLORBUFFER;
//            }
//        }
//        OpenGl.gl.glFinish();
//    }
    
}
