/**
 * Copyright 2005 Jason LaDere
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Originally based on Project64 code.
 *
 */

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package gln64j;

import com.sun.opengl.util.BufferUtil;
import debug.Debug;
import gln64j.rdp.combiners.Combiners;
import gln64j.rdp.Gdp;
import gln64j.rdp.textures.TextureCache;
import java.nio.FloatBuffer;
import java.util.Random;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

/**
 *
 * @author Jason
 */
public class OpenGlGdp {
    
    private static final int SIZEOF_FLOAT = 4;
    private static final int SIZEOF_GLVERTEX = 17 * SIZEOF_FLOAT;
    
    private static class GLSimpleVertex {
        public float x, y, z, w;
        
//        public class VColor {
//            public float r, g, b, a;
//        };
//        public VColor color = new VColor();
//        public VColor secondaryColor = new VColor();
        public float[] color = new float[4]; // r,g,b,a
        public float[] secondaryColor = new float[4]; // r,g,b,a
        
        public float s0, t0, s1, t1;
        
        public float fog;
    };
    
    private static class GLVertex {
        public FloatBuffer vtx; // 4
        public FloatBuffer color; // 4
        public FloatBuffer secondaryColor; // 4
        public FloatBuffer tex0; // 2
        public FloatBuffer tex1; // 2
        public FloatBuffer fog; // 1
        
        public GLVertex() { }
        
        public GLVertex(float x, float y, float z, float w,
                float r1, float g1, float b1, float a1,
                float r2, float g2, float b2, float a2,
                float s0, float t0, float s1, float t1, float f) {
            vtx = FloatBuffer.allocate(4);
            vtx.put(x);
            vtx.put(y);
            vtx.put(z);
            vtx.put(w);
            color = FloatBuffer.allocate(4);
            color.put(r1);
            color.put(g1);
            color.put(b1);
            color.put(a1);
            secondaryColor = FloatBuffer.allocate(4);
            secondaryColor.put(r2);
            secondaryColor.put(g2);
            secondaryColor.put(b2);
            secondaryColor.put(a2);
            tex0 = FloatBuffer.allocate(2);
            tex0.put(s0);
            tex0.put(t0);
            tex1 = FloatBuffer.allocate(2);
            tex1.put(s1);
            tex1.put(t1);
            fog = FloatBuffer.allocate(1);
            fog.put(f);
        }
    };
    
//    public static boolean EXT_fog_coord;		// TNT, GeForce, Rage 128, Radeon
//    public static boolean ATI_texture_env_combine3;	// Radeon
//    public static boolean ATIX_texture_env_route;	// Radeon
//    public static boolean ARB_multitexture;		// TNT, GeForce, Rage 128, Radeon
//    public static boolean ARB_texture_env_combine;	// GeForce, Rage 128, Radeon
//    public static boolean ARB_texture_env_crossbar;	// Radeon (GeForce supports it, but doesn't report it)
//    public static boolean EXT_texture_env_combine;	// TNT, GeForce, Rage 128, Radeon
//    public static boolean EXT_secondary_color;		// GeForce, Radeon
//    public static boolean NV_texture_env_combine4;	// TNT, GeForce
//    public static boolean NV_register_combiners;	// GeForce
//    public static int maxTextureUnits;                  // TNT = 2, GeForce = 2-4, Rage 128 = 2, Radeon = 3-6
    
    public static GL gl;
 
    // used by Gdp
    public static TextureCache cache = new TextureCache();
    public static Combiners combiners = new Combiners();
    public static int screenWidth;
    public static int screenHeight;
    
    // used by GLN64Plugin.processDlist, OpenGl.OGLStart
    public static GLAutoDrawable hDC;
    
    private static int numTriangles;
    private static GLVertex[] vertices = new GLVertex[256];
    private static FloatBuffer bigArray;
    private static float scaleX;
    private static float scaleY;
    private static float zDepth;
    private static int numVertices;
    private static float nearZ;
    private static float vTrans;
    private static float vScale;
    private static int width;
    private static int height;
    private static int heightOffset;
    private static boolean usePolygonStipple;
    private static byte[][][] stipplePattern = new byte[32][8][128];
    private static int lastStipple;
    private static final GLSimpleVertex rect0 = new GLSimpleVertex();
    private static final GLSimpleVertex rect1 = new GLSimpleVertex();
    private static int windowedWidth;
    private static int windowedHeight;
    
    public static void init(GL agl) {
        gl = agl;
            
        for (int i = 0; i < 256; i++)
            vertices[i] = new GLVertex();
        bigArray = BufferUtil.newFloatBuffer(256 * 17);
        for (int i = 0; i < 256; i++) {
            bigArray.position(i * 17);
            vertices[i].vtx = bigArray.slice();
            bigArray.position(i * 17 + 4);
            vertices[i].color = bigArray.slice();
            bigArray.position(i * 17 + 8);
            vertices[i].secondaryColor = bigArray.slice();
            bigArray.position(i * 17 + 12);
            vertices[i].tex0 = bigArray.slice();
            bigArray.position(i * 17 + 14);
            vertices[i].tex1 = bigArray.slice();
            bigArray.position(i * 17 + 16);
            vertices[i].fog = bigArray.slice();
        }

        if (Debug.DEBUG_OGL) System.out.println("GL_NV_register_combiners: " + Combiners.NV_register_combiners);
        Combiners.ARB_multitexture = gl.glGetString(GL.GL_EXTENSIONS).contains("GL_ARB_multitexture");
        if (Debug.DEBUG_OGL) System.out.println("GL_ARB_multitexture: " + Combiners.ARB_multitexture);
        if (Combiners.ARB_multitexture) {
            int[] maxTextureUnits_t = new int[1];
            gl.glGetIntegerv(GL.GL_MAX_TEXTURE_UNITS, maxTextureUnits_t, 0);
            Combiners.maxTextureUnits = StrictMath.min(8, maxTextureUnits_t[0]); // The plugin only supports 8, and 4 is really enough
        }
        Combiners.EXT_fog_coord = gl.glGetString(GL.GL_EXTENSIONS).contains("GL_EXT_fog_coord");
        if (Debug.DEBUG_OGL) System.out.println("GL_EXT_fog_coord: " + Combiners.EXT_fog_coord);
        Combiners.EXT_secondary_color = gl.glGetString(GL.GL_EXTENSIONS).contains("GL_EXT_secondary_color");
        if (Debug.DEBUG_OGL) System.out.println("GL_EXT_secondary_color: " + Combiners.EXT_secondary_color);
        Combiners.ARB_texture_env_combine = gl.glGetString(GL.GL_EXTENSIONS).contains("GL_ARB_texture_env_combine");
        if (Debug.DEBUG_OGL) System.out.println("GL_ARB_texture_env_combine: " + Combiners.ARB_texture_env_combine);
        Combiners.ARB_texture_env_crossbar = gl.glGetString(GL.GL_EXTENSIONS).contains("GL_ARB_texture_env_crossbar");
        if (Debug.DEBUG_OGL) System.out.println("GL_ARB_texture_env_crossbar: " + Combiners.ARB_texture_env_crossbar);
        Combiners.EXT_texture_env_combine = gl.glGetString(GL.GL_EXTENSIONS).contains("GL_EXT_texture_env_combine");
        if (Debug.DEBUG_OGL) System.out.println("GL_EXT_texture_env_combine: " + Combiners.EXT_texture_env_combine);
        Combiners.ATI_texture_env_combine3 = gl.glGetString(GL.GL_EXTENSIONS).contains("GL_ATI_texture_env_combine3");
        if (Debug.DEBUG_OGL) System.out.println("GL_ATI_texture_env_combine3: " + Combiners.ATI_texture_env_combine3);
        if (Debug.DEBUG_OGL) System.out.println("GL_ATIX_texture_env_route: " + Combiners.ATIX_texture_env_route);
        if (Debug.DEBUG_OGL) System.out.println("GL_NV_texture_env_combine4: " + Combiners.NV_texture_env_combine4);

        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();
        bigArray.position(0);
        gl.glVertexPointer(4, GL.GL_FLOAT, SIZEOF_GLVERTEX, bigArray.slice());
        gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
        bigArray.position(4);
        gl.glColorPointer(4, GL.GL_FLOAT, SIZEOF_GLVERTEX, bigArray.slice());
        gl.glEnableClientState(GL.GL_COLOR_ARRAY);
        if (Combiners.EXT_secondary_color) {
            bigArray.position(8);
            gl.glSecondaryColorPointerEXT(3, GL.GL_FLOAT, SIZEOF_GLVERTEX, bigArray.slice());
            gl.glEnableClientState(GL.GL_SECONDARY_COLOR_ARRAY_EXT);
        }
        if (Combiners.ARB_multitexture) {
            gl.glClientActiveTexture(GL.GL_TEXTURE0);
            bigArray.position(12);
            gl.glTexCoordPointer(2, GL.GL_FLOAT, SIZEOF_GLVERTEX, bigArray.slice());
            gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
            gl.glClientActiveTexture(GL.GL_TEXTURE1);
            bigArray.position(14);
            gl.glTexCoordPointer(2, GL.GL_FLOAT, SIZEOF_GLVERTEX, bigArray.slice());
            gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
        } else {
            bigArray.position(12);
            gl.glTexCoordPointer(2, GL.GL_FLOAT, SIZEOF_GLVERTEX, bigArray.slice());
            gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
        }
        if (Combiners.EXT_fog_coord) {
            gl.glFogi(GL.GL_FOG_COORDINATE_SOURCE_EXT, GL.GL_FOG_COORDINATE_EXT);
            gl.glFogi(GL.GL_FOG_MODE, GL.GL_LINEAR);
            gl.glFogf(GL.GL_FOG_START, 0.0f);
            gl.glFogf(GL.GL_FOG_END, 255.0f);
            bigArray.position(16);
            gl.glFogCoordPointerEXT(GL.GL_FLOAT, SIZEOF_GLVERTEX, bigArray.slice());
            gl.glEnableClientState(GL.GL_FOG_COORDINATE_ARRAY_EXT);
        }
        gl.glPolygonOffset(-3.0f, -3.0f);
        gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        gl.glClear(GL.GL_COLOR_BUFFER_BIT);
        
        Random rand = new Random();
        for (int i = 0; i < 32; i++) {
            for (int j = 0; j < 8; j++) {
                for (int k = 0; k < 128; k++) {
                    stipplePattern[i][j][k] = (byte) (((i > (rand.nextInt() >>> 10) ? 1 : 0) << 7) |
                            ((i > (rand.nextInt() >>> 10) ? 1 : 0) << 6) |
                            ((i > (rand.nextInt() >>> 10) ? 1 : 0) << 5) |
                            ((i > (rand.nextInt() >>> 10) ? 1 : 0) << 4) |
                            ((i > (rand.nextInt() >>> 10) ? 1 : 0) << 3) |
                            ((i > (rand.nextInt() >>> 10) ? 1 : 0) << 2) |
                            ((i > (rand.nextInt() >>> 10) ? 1 : 0) << 1) |
                            ((i > (rand.nextInt() >>> 10) ? 1 : 0) << 0));
                }
            }
        }
        usePolygonStipple = false;
        cache.init(gl, GLN64jPlugin.RDRAM, GLN64jPlugin.TMEM, Combiners.maxTextureUnits, Combiners.ARB_multitexture);
        combiners.init();
        OGL_UpdateScale();
    }
    
    // called by OpenGl.config
    public static void config() {
        cache.config(32 * 1048576, 1);
        windowedWidth = 640;
        windowedHeight = 480;
    }
    
    // called by GLN64Plugin.romOpen
    public static void OGL_ResizeWindow() {
        width = windowedWidth;
        height = windowedHeight;
    }
    
    // called by OpenGl.display
    public static void OGL_UpdateScale() {
        scaleX = width / (float)screenWidth;
        scaleY = height / (float)screenHeight;
    }
    
    // called by OpenGl.display
    public static void VI_UpdateSize() {
        float xScale = (GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_X_SCALE]&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
//        float xOffset = ((GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_X_SCALE]>>16)&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
        
        float yScale = (GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_Y_SCALE]&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
//        float yOffset = ((GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_X_SCALE]>>16)&Gbi.SR_MASK_12) * Gbi.FIXED2FLOATRECIP10;
        
        int hEnd = GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_H_START]&Gbi.SR_MASK_10;
        int hStart = (GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_H_START]>>16)&Gbi.SR_MASK_10;
        
        // These are in half-lines, so shift an extra bit
        int vEnd = (GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_V_START]>>1)&Gbi.SR_MASK_9;
        int vStart = (GLN64jPlugin.REG.VI_Registers[GLN64jPlugin.REG.VI_V_START]>>17)&Gbi.SR_MASK_9;
        
        screenWidth = (int)((hEnd - hStart) * xScale);
        screenHeight = (int)((vEnd - vStart) * yScale * 1.0126582f);
        
        if (screenWidth == 0)
            screenWidth = 320;
        if (screenHeight == 0)
            screenHeight = 240;
    }
    
    // called by GLN64jPlugin.updateScreen
    public static void viUpdateScreen() {
//        if (gl==null)
//            return;
//        gl.glFinish();
        if ((Rsp.gdp.changed & Gdp.CHANGED_COLORBUFFER) != 0) {
            hDC.swapBuffers();
            Rsp.gdp.changed &= ~Gdp.CHANGED_COLORBUFFER;
        }
//        gl.glFinish();
    }
    
    // called by Gsp.viewport
    public static void setView(float trans, float scale) {
        vTrans = trans;
        vScale = scale;
        nearZ = trans - scale;
        Rsp.gdp.changed |= Gdp.CHANGED_VIEWPORT;
    }
    
    // called by Gdp.setPrimDepth
    public static void setZDepth(int z) {
        zDepth = StrictMath.min(1.0f, StrictMath.max(0.0f, ((z * Gbi.FIXED2FLOATRECIP15) - vTrans) / vScale));
    }
    
    // called by Gdp.setDepthImage
    public static void OGL_ClearDepthBuffer(boolean depthUpdate) {
        OpenGl.OGL_GspUpdateStates();
        
        OGL_GdpUpdateStates();
        gl.glDepthMask(true);
        gl.glClear(GL.GL_DEPTH_BUFFER_BIT);
        gl.glDepthMask(depthUpdate);
    }
    
    // called by Gdp.fillRectangle
    public static void OGL_ClearColorBuffer(float[] color) {
        gl.glClearColor(color[0], color[1], color[2], color[3]);
        gl.glClear(GL.GL_COLOR_BUFFER_BIT);
    }
    
    // called by Gdp.fillRectangle
    // uses screenWidth, screenHeight, width, height, heightOffset, zDepth, nearZ
    public static void OGL_DrawRect(int ulx, int uly, int lrx, int lry, float[] color, int depthSource) {
        OpenGl.OGL_GspUpdateStates();
        
        OGL_GdpUpdateStates();
        
        boolean culling = gl.glIsEnabled(GL.GL_CULL_FACE);
        gl.glDisable(GL.GL_SCISSOR_TEST);
        gl.glDisable(GL.GL_CULL_FACE);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glOrtho(0, screenWidth, screenHeight, 0, 1.0f, -1.0f);
        gl.glViewport(0, heightOffset, width, height);
        if (Debug.WIREFRAME)
            gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE); //TMP
        gl.glDepthRange(0.0f, 1.0f);
        gl.glColor4f(color[0], color[1], color[2], color[3]);
        
        gl.glBegin(GL.GL_QUADS);
        gl.glVertex4f(ulx, uly, (depthSource == Gbi.G_ZS_PRIM) ? zDepth : nearZ, 1.0f);
        gl.glVertex4f(lrx, uly, (depthSource == Gbi.G_ZS_PRIM) ? zDepth : nearZ, 1.0f);
        gl.glVertex4f(lrx, lry, (depthSource == Gbi.G_ZS_PRIM) ? zDepth : nearZ, 1.0f);
        gl.glVertex4f(ulx, lry, (depthSource == Gbi.G_ZS_PRIM) ? zDepth : nearZ, 1.0f);
        gl.glEnd();
        
        gl.glLoadIdentity();
        
//        OpenGl.OGL_UpdateCullFace();
        if (culling)
            gl.glEnable(GL.GL_CULL_FACE);
        
//        OGL_UpdateViewport();
        gl.glEnable(GL.GL_SCISSOR_TEST);
    }
    
    // called by TextureCache.drawTextureRectangle
    // uses screenWidth, screenHeight, width, height, heightOffset, zDepth, nearZ, rect0, rect1
    public static void OGL_DrawTexturedRect(float ulx, float uly, float lrx, float lry, float uls, float ult, float lrs, float lrt, boolean flip, int depthSource, int cycleType) {
        OpenGl.OGL_GspUpdateStates();
        
        OGL_GdpUpdateStates();
        
        rect0.x = ulx;
        rect0.y = uly;
        rect0.z = depthSource == Gbi.G_ZS_PRIM ? zDepth : nearZ;
        rect0.w = 1.0f;
        rect0.color[0] = 1.0f;
        rect0.color[1] = 1.0f;
        rect0.color[2] = 1.0f;
        rect0.color[3] = 0.0f;
        rect0.secondaryColor[0] = 1.0f;
        rect0.secondaryColor[1] = 1.0f;
        rect0.secondaryColor[2] = 1.0f;
        rect0.secondaryColor[3] = 1.0f;
        rect0.s0 = uls;
        rect0.t0 = ult;
        rect0.s1 = uls;
        rect0.t1 = ult;
        rect0.fog = 0.0f;
        
        rect1.x = lrx;
        rect1.y = lry;
        rect1.z = depthSource == Gbi.G_ZS_PRIM ? zDepth : nearZ;
        rect1.w = 1.0f;
        rect1.color[0] = 1.0f;
        rect1.color[1] = 1.0f;
        rect1.color[2] = 1.0f;
        rect1.color[3] = 0.0f;
        rect1.secondaryColor[0] = 1.0f;
        rect1.secondaryColor[1] = 1.0f;
        rect1.secondaryColor[2] = 1.0f;
        rect1.secondaryColor[3] = 1.0f;
        rect1.s0 = lrs;
        rect1.t0 = lrt;
        rect1.s1 = lrs;
        rect1.t1 = lrt;
        rect1.fog = 0.0f;
        
        boolean culling = gl.glIsEnabled(GL.GL_CULL_FACE);
        gl.glDisable(GL.GL_CULL_FACE);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glOrtho(0, screenWidth, screenHeight, 0, 1.0f, -1.0f); // left, right, bottom, top, near, far
        gl.glViewport(0, heightOffset, width, height); // x, y, width, height
        if (Debug.WIREFRAME)
            gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE); //TMP
        
        if (combiners.usesT0) {
            rect0.s0 = rect0.s0 * cache.current[0].shiftScaleS - cache.textureTile[0].fuls;
            rect0.t0 = rect0.t0 * cache.current[0].shiftScaleT - cache.textureTile[0].fult;
            rect1.s0 = (rect1.s0 + 1.0f) * cache.current[0].shiftScaleS - cache.textureTile[0].fuls;
            rect1.t0 = (rect1.t0 + 1.0f) * cache.current[0].shiftScaleT - cache.textureTile[0].fult;
            
            if ((cache.current[0].maskS != 0) && (StrictMath.IEEEremainder(rect0.s0, cache.current[0].width) == 0.0f) && (cache.current[0].mirrorS == 0)) {
                rect1.s0 -= rect0.s0;
                rect0.s0 = 0.0f;
            }
            if ((cache.current[0].maskT != 0) && (StrictMath.IEEEremainder(rect0.t0, cache.current[0].height) == 0.0f) && (cache.current[0].mirrorT == 0)) {
                rect1.t0 -= rect0.t0;
                rect0.t0 = 0.0f;
            }
            
            if (Combiners.ARB_multitexture)
                gl.glActiveTexture(GL.GL_TEXTURE0);
            
            if ((rect0.s0 >= 0.0f) && (rect1.s0 <= cache.current[0].width)) {
                gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE);
            }
            if ((rect0.t0 >= 0.0f) && (rect1.t0 <= cache.current[0].height)) {
                gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE);
            }
            
            rect0.s0 *= cache.current[0].scaleS;
            rect0.t0 *= cache.current[0].scaleT;
            rect1.s0 *= cache.current[0].scaleS;
            rect1.t0 *= cache.current[0].scaleT;
        }
        
        if (combiners.usesT1 && Combiners.ARB_multitexture) {
            rect0.s1 = rect0.s1 * cache.current[1].shiftScaleS - cache.textureTile[1].fuls;
            rect0.t1 = rect0.t1 * cache.current[1].shiftScaleT - cache.textureTile[1].fult;
            rect1.s1 = (rect1.s1 + 1.0f) * cache.current[1].shiftScaleS - cache.textureTile[1].fuls;
            rect1.t1 = (rect1.t1 + 1.0f) * cache.current[1].shiftScaleT - cache.textureTile[1].fult;
            
            if ((cache.current[1].maskS != 0) && (StrictMath.IEEEremainder(rect0.s1, cache.current[1].width) == 0.0f) && (cache.current[1].mirrorS == 0)) {
                rect1.s1 -= rect0.s1;
                rect0.s1 = 0.0f;
            }
            if ((cache.current[1].maskT != 0) && (StrictMath.IEEEremainder(rect0.t1, cache.current[1].height) == 0.0f) && (cache.current[1].mirrorT == 0)) {
                rect1.t1 -= rect0.t1;
                rect0.t1 = 0.0f;
            }
            
            gl.glActiveTexture(GL.GL_TEXTURE1);
            
            if ((rect0.s1 == 0.0f) && (rect1.s1 <= cache.current[1].width)) {
                gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE);
            }
            if ((rect0.t1 == 0.0f) && (rect1.t1 <= cache.current[1].height)) {
                gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE);
            }
            
            rect0.s1 *= cache.current[1].scaleS;
            rect0.t1 *= cache.current[1].scaleT;
            rect1.s1 *= cache.current[1].scaleS;
            rect1.t1 *= cache.current[1].scaleT;
        }
        
//        if ((cycleType == Gbi.G_CYC_COPY) && !forceBilinear) {
        if (cycleType == Gbi.G_CYC_COPY) {
            if (Combiners.ARB_multitexture)
                gl.glActiveTexture(GL.GL_TEXTURE0);
            
            gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST);
            gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST);
        }
        
        combiners.setConstant(rect0.color, combiners.vertex.color, combiners.vertex.alpha);
        
        if (Combiners.EXT_secondary_color)
            combiners.setConstant(rect0.secondaryColor, combiners.vertex.secondaryColor, combiners.vertex.alpha);
        
        gl.glBegin(GL.GL_QUADS);
//        gl.glColor4f(rect0.color.r, rect0.color.g, rect0.color.b, rect0.color.a);
        gl.glColor4f(rect0.color[0], rect0.color[1], rect0.color[2], rect0.color[3]);
        if (Combiners.EXT_secondary_color)
//            gl.glSecondaryColor3fEXT(rect0.secondaryColor.r, rect0.secondaryColor.g, rect0.secondaryColor.b);
            gl.glSecondaryColor3fEXT(rect0.secondaryColor[0], rect0.secondaryColor[1], rect0.secondaryColor[2]);
        
        if (Combiners.ARB_multitexture) {
            gl.glMultiTexCoord2f(GL.GL_TEXTURE0, rect0.s0, rect0.t0);
            gl.glMultiTexCoord2f(GL.GL_TEXTURE1, rect0.s1, rect0.t1);
            gl.glVertex4f(rect0.x, rect0.y, rect0.z, 1.0f);
            
            gl.glMultiTexCoord2f(GL.GL_TEXTURE0, rect1.s0, rect0.t0);
            gl.glMultiTexCoord2f(GL.GL_TEXTURE1, rect1.s1, rect0.t1);
            gl.glVertex4f(rect1.x, rect0.y, rect0.z, 1.0f);
            
            gl.glMultiTexCoord2f(GL.GL_TEXTURE0, rect1.s0, rect1.t0);
            gl.glMultiTexCoord2f(GL.GL_TEXTURE1, rect1.s1, rect1.t1);
            gl.glVertex4f(rect1.x, rect1.y, rect0.z, 1.0f);
            
            gl.glMultiTexCoord2f(GL.GL_TEXTURE0, rect0.s0, rect1.t0);
            gl.glMultiTexCoord2f(GL.GL_TEXTURE1, rect0.s1, rect1.t1);
            gl.glVertex4f(rect0.x, rect1.y, rect0.z, 1.0f);
        } else {
            gl.glTexCoord2f(rect0.s0, rect0.t0);
            gl.glVertex4f(rect0.x, rect0.y, rect0.z, 1.0f);
            
            if (flip)
                gl.glTexCoord2f(rect0.s0, rect1.t0);
            else
                gl.glTexCoord2f(rect1.s0, rect0.t0);
            gl.glVertex4f(rect1.x, rect0.y, rect0.z, 1.0f);
            
            gl.glTexCoord2f(rect1.s0, rect1.t0);
            gl.glVertex4f(rect1.x, rect1.y, rect0.z, 1.0f);
            
            if (flip)
                gl.glTexCoord2f(rect1.s0, rect0.t0);
            else
                gl.glTexCoord2f(rect0.s0, rect1.t0);
            gl.glVertex4f(rect0.x, rect1.y, rect0.z, 1.0f);
        }
        gl.glEnd();
        
        gl.glLoadIdentity();
        
//        OpenGl.OGL_UpdateCullFace();
        if (culling)
            gl.glEnable(GL.GL_CULL_FACE);
        
//        OGL_UpdateViewport();
    }
    
    // called by Gsp.gSPLine3D
    // uses scaleX, combiners
    public static void OGL_DrawLine(float[] vtx1, float[] c1, float[] vtx2, float[] c2, float width) {
        if (Rsp.gsp.changed != 0 || Rsp.gdp.changed != 0) {
            OpenGl.OGL_GspUpdateStates();
            OGL_GdpUpdateStates();
        }
        
        final int VTX = 0;
        final int CLR = 1;
        float[][][] v = {{vtx1, c1}, {vtx2, c2}};
        
        float[] color = new float[4];
        
//        if (Rsp.gsp.changed != 0 || Rsp.gdp.changed != 0) {
//            OpenGl.OGL_GspUpdateStates();
//            OGL_GdpUpdateStates();
//        }
        
        gl.glLineWidth(width * scaleX);
        
        gl.glBegin(GL.GL_LINES);
        for (int i = 0; i < 2; i++) {
            float[][] spvert = v[i];
            color[0] = spvert[CLR][0];
            color[1] = spvert[CLR][1];
            color[2] = spvert[CLR][2];
            color[3] = spvert[CLR][3];
            combiners.setConstant(color, combiners.vertex.color, combiners.vertex.alpha);
            gl.glColor4fv(color, 0);
            
            if (Combiners.EXT_secondary_color) {
                color[0] = spvert[CLR][0];
                color[1] = spvert[CLR][1];
                color[2] = spvert[CLR][2];
                color[3] = spvert[CLR][3];
                combiners.setConstant(color, combiners.vertex.secondaryColor, combiners.vertex.alpha);
                gl.glSecondaryColor3fvEXT(color, 0);
            }
            
            gl.glVertex4f(spvert[VTX][0], spvert[VTX][1], spvert[VTX][2], spvert[VTX][3]);
        }
        gl.glEnd();
    }
    
    // called by Gsp.gSPTriangle
    // uses vertices, numVertices, numTriangles, combiners, cache
    public static void OGL_AddTriangle(float[] vtx1, float[] c1, float[] tex1, float[] vtx2, float[] c2, float[] tex2, float[] vtx3, float[] c3, float[] tex3) {
        if (Rsp.gsp.changed != 0 || Rsp.gdp.changed != 0) {
            OpenGl.OGL_GspUpdateStates();
            OGL_GdpUpdateStates();
        }
        
        final int VTX = 0;
        final int CLR = 1;
        final int TEX = 2;
        float[][][] v = {{vtx1, c1, tex1}, {vtx2, c2, tex2}, {vtx3, c3, tex3}};
        
//        if (Rsp.gsp.changed != 0 || Rsp.gdp.changed != 0) {
//            OpenGl.OGL_GspUpdateStates();
//            OGL_GdpUpdateStates();
//        }

        for (int i = 0; i < 3; i++) {
            float[][] spvert = v[i];
            GLVertex vertex = vertices[numVertices];
            vertex.vtx.put(0, spvert[VTX][0]);
            vertex.vtx.put(1, spvert[VTX][1]);
//            vertex.vtx.put(2, Rsp.gdp.otherMode.depthSource == Gbi.G_ZS_PRIM ? zDepth * spvert[VTX][3] : spvert[VTX][2]);
            vertex.vtx.put(2, Gdp.RDP_GETOM_Z_SOURCE_SEL(Rsp.gdp.otherMode) == Gbi.G_ZS_PRIM ? zDepth * spvert[VTX][3] : spvert[VTX][2]);
            vertex.vtx.put(3, spvert[VTX][3]);
            
            vertex.color.put(0, spvert[CLR][0]);
            vertex.color.put(1, spvert[CLR][1]);
            vertex.color.put(2, spvert[CLR][2]);
            vertex.color.put(3, spvert[CLR][3]);
            combiners.setConstant(vertex.color, combiners.vertex.color, combiners.vertex.alpha);
            
            if (Combiners.EXT_secondary_color) {
                vertex.secondaryColor.put(0, 0.0f);
                vertex.secondaryColor.put(0, 0.0f);
                vertex.secondaryColor.put(0, 0.0f);
                vertex.secondaryColor.put(0, 1.0f);
                combiners.setConstant(vertex.secondaryColor, combiners.vertex.secondaryColor, Combiners.ONE);
            }
            
            if (combiners.usesT0) {
                vertex.tex0.put(0, cache.getTexS(0, spvert[TEX][0]));
                vertex.tex0.put(1, cache.getTexT(0, spvert[TEX][1]));
            }
            if (combiners.usesT1) {
                vertex.tex1.put(0, cache.getTexS(1, spvert[TEX][0]));
                vertex.tex1.put(1, cache.getTexT(1, spvert[TEX][1]));
            }
            numVertices++;
        }
        
        numTriangles++;
        
        if (numVertices >= 255)
            OGL_DrawTriangles();
    }
    
    // called by Gsp.gSPFlushTriangles
    public static void OGL_DrawTriangles() {
        if (numTriangles < 1)
            return;
        stipple();
        gl.glDrawArrays(GL.GL_TRIANGLES, 0, numVertices);
        numTriangles = numVertices = 0;
    }
    
    // Private Methods /////////////////////////////////////////////////////////
    
    private static void stipple() {
//        if (usePolygonStipple && (Rsp.gdp.otherMode.alphaCompare == Gbi.G_AC_DITHER) && (Rsp.gdp.otherMode.alphaCvgSel == 0)) {
        if (usePolygonStipple && (Gdp.RDP_GETOM_ALPHA_COMPARE_EN(Rsp.gdp.otherMode) == Gbi.G_AC_DITHER) && (Gdp.RDP_GETOM_ALPHA_CVG_SELECT(Rsp.gdp.otherMode) == 0)) {
            lastStipple = (lastStipple + 1) & 0x7;
            gl.glPolygonStipple(stipplePattern[(int)(Combiners.envColor[3] * 255.0f) >> 3][lastStipple], 0);
        }
    }
    
//    // called by TextureCache.activateTexture
//    public static void activateTexture(int t, CachedTexture texture, int filter) {
//        if (ARB_multitexture)
//            gl.glActiveTexture(GL.GL_TEXTURE0 + t);
//        
//        // Bind the cached texture
//        gl.glBindTexture(GL.GL_TEXTURE_2D, texture.glName[0]);
//        
//        // Set filter mode. Almost always bilinear, but check anyways
////        if ((Rsp.gdp.otherMode.textureFilter == Gbi.G_TF_BILERP) || (Rsp.gdp.otherMode.textureFilter == Gbi.G_TF_AVERAGE) || (forceBilinear)) {
//        if ((filter == Gbi.G_TF_BILERP) || (filter == Gbi.G_TF_AVERAGE)) {
//            gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
//            gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);
//        } else {
//            gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST);
//            gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST);
//        }
//        
//        // Set clamping modes
//        gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, texture.clampS!=0 ? GL.GL_CLAMP_TO_EDGE : GL.GL_REPEAT);
//        gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, texture.clampT!=0 ? GL.GL_CLAMP_TO_EDGE : GL.GL_REPEAT);
//        
////        texture.lastDList = Rsp.gsp.DList;
//    }
    
//    public static void OGL_UpdateViewport() {
//        
//    }
    
    private static void OGL_GdpUpdateStates() {
        if ((Rsp.gdp.changed & Gdp.CHANGED_RENDERMODE) != 0) {
//            if (Rsp.gdp.otherMode.depthCompare != 0)
            if (Gdp.RDP_GETOM_Z_COMPARE_EN(Rsp.gdp.otherMode) != 0)
                gl.glDepthFunc(GL.GL_LEQUAL);
            else
                gl.glDepthFunc(GL.GL_ALWAYS);
            
//            OGL_UpdateDepthUpdate();
//            if (Rsp.gdp.otherMode.depthUpdate != 0)
            if (Gdp.RDP_GETOM_Z_UPDATE_EN(Rsp.gdp.otherMode) != 0)
                gl.glDepthMask(true);
            else
                gl.glDepthMask(false);
            
//            if (Rsp.gdp.otherMode.depthMode == Gbi.ZMODE_DEC)
            if (Gdp.RDP_GETOM_Z_MODE(Rsp.gdp.otherMode) == Gbi.ZMODE_DEC)
                gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
            else {
                gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);
            }
        }
        
        if ((Rsp.gdp.changed & Gdp.CHANGED_ALPHACOMPARE) != 0 || (Rsp.gdp.changed & Gdp.CHANGED_RENDERMODE) != 0) {
            // Enable alpha test for threshold mode
//            if ((Rsp.gdp.otherMode.alphaCompare == Gbi.G_AC_THRESHOLD) && (Rsp.gdp.otherMode.alphaCvgSel == 0)) {
            if ((Gdp.RDP_GETOM_ALPHA_COMPARE_EN(Rsp.gdp.otherMode) == Gbi.G_AC_THRESHOLD) && (Gdp.RDP_GETOM_ALPHA_CVG_SELECT(Rsp.gdp.otherMode) == 0)) {
                gl.glEnable(GL.GL_ALPHA_TEST);
                
                gl.glAlphaFunc((Rsp.gdp.blendColor[3] > 0.0f) ? GL.GL_GEQUAL : GL.GL_GREATER, Rsp.gdp.blendColor[3]);
            }
            // Used in TEX_EDGE and similar render modes
//            else if (Rsp.gdp.otherMode.cvgXAlpha != 0) {
            else if (Gdp.RDP_GETOM_CVG_TIMES_ALPHA(Rsp.gdp.otherMode) != 0) {
                gl.glEnable(GL.GL_ALPHA_TEST);
                
                // Arbitrary number -- gives nice results though
                gl.glAlphaFunc(GL.GL_GEQUAL, 0.5f);
            } else {
                gl.glDisable(GL.GL_ALPHA_TEST);
            }
            
//            if (usePolygonStipple && (Rsp.gdp.otherMode.alphaCompare == Gbi.G_AC_DITHER) && (Rsp.gdp.otherMode.alphaCvgSel == 0))
            if (usePolygonStipple && (Gdp.RDP_GETOM_ALPHA_COMPARE_EN(Rsp.gdp.otherMode) == Gbi.G_AC_DITHER) && (Gdp.RDP_GETOM_ALPHA_CVG_SELECT(Rsp.gdp.otherMode) == 0))
                gl.glEnable(GL.GL_POLYGON_STIPPLE);
            else
                gl.glDisable(GL.GL_POLYGON_STIPPLE);
        }
        
        if ((Rsp.gdp.changed & Gdp.CHANGED_SCISSOR) != 0) {
            gl.glScissor((int)(Rsp.gdp.scissor.ulx * scaleX), (int)((screenHeight - Rsp.gdp.scissor.lry) * scaleY + heightOffset),
                    (int)((Rsp.gdp.scissor.lrx - Rsp.gdp.scissor.ulx) * scaleX), (int)((Rsp.gdp.scissor.lry - Rsp.gdp.scissor.uly) * scaleY));
        }
        
//        if ((Rsp.gdp.changed & Gdp.CHANGED_VIEWPORT) != 0) {
//            OGL_UpdateViewport();
//        }
        
        if ((Rsp.gdp.changed & Gdp.CHANGED_COMBINE) != 0 || (Rsp.gdp.changed & Gdp.CHANGED_CYCLETYPE) != 0) {
//            if (Rsp.gdp.otherMode.cycleType == Gbi.G_CYC_COPY) {
            if (Gdp.RDP_GETOM_CYCLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_CYC_COPY) {
//                combiners.setCombine(gl, Rsp.gdp.otherMode.cycleType == Gbi.G_CYC_2CYCLE, combiners.encodeCombineMode(
                combiners.setCombine(gl, false, combiners.encodeCombineMode(
                        Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_TEXEL0,
                        Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_TEXEL0,
                        Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_TEXEL0,
                        Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_TEXEL0));
//            } else if (Rsp.gdp.otherMode.cycleType == Gbi.G_CYC_FILL) {
                } else if (Gdp.RDP_GETOM_CYCLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_CYC_FILL) {
//                combiners.setCombine(gl, Rsp.gdp.otherMode.cycleType == Gbi.G_CYC_2CYCLE, combiners.encodeCombineMode(
                combiners.setCombine(gl, false, combiners.encodeCombineMode(
                        Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_SHADE,
                        Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_1,
                        Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_0, Combiners.G_CCMUX_SHADE,
                        Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_0, Combiners.G_ACMUX_1));
            } else {
//                combiners.setCombine(gl, Rsp.gdp.otherMode.cycleType == Gbi.G_CYC_2CYCLE, combiners.combine.getMux());
                combiners.setCombine(gl, Gdp.RDP_GETOM_CYCLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_CYC_2CYCLE, combiners.combine.getMux());
            }
            Rsp.gdp.changed |= Gdp.CHANGED_COMBINE_COLORS;
        }
        
        if ((Rsp.gdp.changed & Gdp.CHANGED_COMBINE_COLORS) != 0) {
            combiners.updateCombineColors(gl);
            Rsp.gdp.changed &= ~Gdp.CHANGED_COMBINE_COLORS;
        }
        
        if ((Rsp.gdp.changed & Gdp.CHANGED_TEXTURE) != 0 || (Rsp.gdp.changed & Gdp.CHANGED_TILE) != 0 || (cache.changed & TextureCache.CHANGED_TMEM) != 0) {
            combiners.beginTextureUpdate(gl);
            
            if (combiners.usesT0) {
//                cache.update(Rsp.gdp.otherMode.textureLUT == Gbi.G_TT_IA16, 0, scaleX, scaleY, Rsp.gdp.otherMode.textureFilter);
//                cache.update(Rsp.gdp.otherMode.textureLUT == Gbi.G_TT_IA16, 0, scaleX, scaleY, (Rsp.gdp.otherMode.textureFilter == Gbi.G_TF_BILERP) || (Rsp.gdp.otherMode.textureFilter == Gbi.G_TF_AVERAGE));
                cache.update(Gdp.RDP_GETOM_TLUT_TYPE(Rsp.gdp.otherMode) == Gbi.G_TT_IA16, 0, scaleX, scaleY, (Gdp.RDP_GETOM_SAMPLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_TF_BILERP) || (Gdp.RDP_GETOM_SAMPLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_TF_AVERAGE));
                
                Rsp.gdp.changed &= ~Gdp.CHANGED_TEXTURE;
                Rsp.gdp.changed &= ~Gdp.CHANGED_TILE;
                cache.changed &= ~TextureCache.CHANGED_TMEM;
            } else {
                cache.activateDummy(0);
            }
            
            //NOTE: The following block does not appear to work if
            // combiner uses T0 but not T1
            if (combiners.usesT1) {
//                cache.update(Rsp.gdp.otherMode.textureLUT == Gbi.G_TT_IA16, 1, scaleX, scaleY, Rsp.gdp.otherMode.textureFilter);
//                cache.update(Rsp.gdp.otherMode.textureLUT == Gbi.G_TT_IA16, 1, scaleX, scaleY, (Rsp.gdp.otherMode.textureFilter == Gbi.G_TF_BILERP) || (Rsp.gdp.otherMode.textureFilter == Gbi.G_TF_AVERAGE));
                cache.update(Gdp.RDP_GETOM_TLUT_TYPE(Rsp.gdp.otherMode) == Gbi.G_TT_IA16, 1, scaleX, scaleY, (Gdp.RDP_GETOM_SAMPLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_TF_BILERP) || (Gdp.RDP_GETOM_SAMPLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_TF_AVERAGE));
                
                Rsp.gdp.changed &= ~Gdp.CHANGED_TEXTURE;
                Rsp.gdp.changed &= ~Gdp.CHANGED_TILE;
                cache.changed &= ~TextureCache.CHANGED_TMEM;
            } else {
                cache.activateDummy(1);
            }
            
            combiners.endTextureUpdate(gl);
        }
        
//        if ((Rsp.gdp.changed & Gdp.CHANGED_FOGCOLOR) != 0 && fog)
//            gl.glFogfv(GL.GL_FOG_COLOR, Rsp.gdp.fogColor, 0);
        
        if ((Rsp.gdp.changed & Gdp.CHANGED_RENDERMODE) != 0 || (Rsp.gdp.changed & Gdp.CHANGED_CYCLETYPE) != 0) {
//            if ((Rsp.gdp.otherMode.forceBlender != 0) &&
//                    (Rsp.gdp.otherMode.cycleType != Gbi.G_CYC_COPY) &&
//                    (Rsp.gdp.otherMode.cycleType != Gbi.G_CYC_FILL) &&
//                    (Rsp.gdp.otherMode.alphaCvgSel == 0)) {
            if ((Gdp.RDP_GETOM_FORCE_BLEND(Rsp.gdp.otherMode) != 0) &&
                    (Gdp.RDP_GETOM_CYCLE_TYPE(Rsp.gdp.otherMode) != Gbi.G_CYC_COPY) &&
                    (Gdp.RDP_GETOM_CYCLE_TYPE(Rsp.gdp.otherMode) != Gbi.G_CYC_FILL) &&
                    (Gdp.RDP_GETOM_ALPHA_CVG_SELECT(Rsp.gdp.otherMode) == 0)) {
                gl.glEnable(GL.GL_BLEND);
                
//                switch (Rsp.gdp.otherMode.getL() >> 16) {
                switch (Rsp.gdp.otherMode.w1 >> 16) {
                    case 0x0448: // Add
                    case 0x055A:
                        gl.glBlendFunc(GL.GL_ONE, GL.GL_ONE);
                        break;
                    case 0x0C08: // 1080 Sky
                    case 0x0F0A: // Used LOTS of places
                        gl.glBlendFunc(GL.GL_ONE, GL.GL_ZERO);
                        break;
                    case 0xC810: // Blends fog
                    case 0xC811: // Blends fog
                    case 0x0C18: // Standard interpolated blend
                    case 0x0C19: // Used for antialiasing
                    case 0x0050: // Standard interpolated blend
                    case 0x0055: // Used for antialiasing
                        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                        break;
                    case 0x0FA5: // Seems to be doing just blend color - maybe combiner can be used for this?
                    case 0x5055: // Used in Paper Mario intro, I'm not sure if this is right...
                        gl.glBlendFunc(GL.GL_ZERO, GL.GL_ONE);
                        break;
                    default:
                        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                        break;
                }
            } else {
                gl.glDisable(GL.GL_BLEND);
            }
            
//            if (Rsp.gdp.otherMode.cycleType == Gbi.G_CYC_FILL) {
            if (Gdp.RDP_GETOM_CYCLE_TYPE(Rsp.gdp.otherMode) == Gbi.G_CYC_FILL) {
                gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                gl.glEnable(GL.GL_BLEND);
            }
        }
        
        cache.changed &= TextureCache.CHANGED_TMEM;
        Rsp.gdp.changed &= Gdp.CHANGED_TILE;
        Rsp.gdp.changed &= Gdp.CHANGED_TEXTURE;
    }

}
